/*
This file was generated by Dashcode from Apple Inc.
You may edit this file to customize your Dashboard widget.
*/

body {
    margin: 0px;
}

#frontImg {
    position: absolute;
    top: 0px;
    left: 0px;
    bottom: 0px;
    right: 0px;
    width: auto;
    height: auto;
}

#info {
    position: absolute;
    right: 15px;
    bottom: 18px;
    width: 13px;
    height: 13px;
}

#back {
    display: none;
}

#backImg {
    position: absolute;
    top: 0px;
    left: 0px;
    right: 0px;
    width: auto;
    bottom: 0px;
    height: auto;
}

#done {
    position: absolute;
    right: 21px;
    bottom: 21px;
    margin-left: auto;
    margin-top: auto;
}

#radiobutton7 {
}

#timeStyleSSdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    left: 161px;
    top: 69px;
    width: 186px;
}

#timeStyleHHMMSSdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    left: 161px;
    top: 29px;
    width: 186px;
}

#timeStyleSS {
}

#timeStyleHHMMdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    left: 161px;
    top: 49px;
    width: 186px;
}

#hourStyle12div {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    width: 103px;
    top: 29px;
    left: 356px;
}

#hourStyle24div {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    width: 103px;
    top: 49px;
    left: 356px;
}

#dateStyleMMDDYYdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    width: 120px;
    left: 30px;
    top: 29px;
}

#dateStyleDDMMYYdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    width: 120px;
    left: 30px;
    top: 49px;
}

#dateStyleYYMMDDdiv {
    position: absolute;
    font-family: HelveticaNeue;
    font-size: 10pt;
    width: 120px;
    left: 30px;
    top: 69px;
}

#errorDiv {
    position: absolute;
    top: 30px;
    left: 30px;
    bottom: 30px;
    right: 30px;

    padding: 40px;
    text-align: center;
    font-family: HelveticaNeue;
    font-size: 16pt;
    font-weight: bold;
    color: white;
    background: black;
    display: none;
}

