/* Dali Clock - a melting digital clock for Palm WebOS.
 * Copyright (c) 1991-2009 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

#clockbg { position:absolute; left:0px; top:0px; }
#canvas  { position:absolute; left:0px; top:0px; }

.prefslabel { color: #1F75BF; font-size: 14px; text-transform: uppercase;
	      margin: 0.2em 0 1em 1.5em; }
