<!-- Dali Clock - a melting digital clock for Palm WebOS.
     Copyright (c) 1991-2009 Jamie Zawinski <jwz@jwz.org>

     Permission to use, copy, modify, distribute, and sell this
     software and its documentation for any purpose is hereby granted
     without fee, provided that the above copyright notice appear in
     all copies and that both that copyright notice and this
     permission notice appear in supporting documentation.  No
     representations are made about the suitability of this software
     for any purpose.  It is provided "as is" without express or
     implied warranty.
  -->

<div id="main" class="palm-hasheader">
 <div class="palm-header">Dali Clock</div>

 <div class="palm-group unlabeled" id="inputwapper">
  <div class="palm-list">
   <div class="palm-row first">
    <div id="timeModeSelector" x-mojo-element="ListSelector"></div>
   </div>
   <div class="palm-row last">
    <div id="dateModeSelector" x-mojo-element="ListSelector"></div>
   </div>
  </div>
 </div>

 <div id="twelveHourRadio" x-mojo-element="RadioButton"></div>

 <div class="palm-group unlabeled" id="inputwapper">
  <div class="palm-list">

   <div class="palm-row first">
    <div class="prefslabel">Target Frame Rate</div>
    <div id="fpsSlider" x-mojo-element="Slider"></div>
   </div>

   <div class="palm-row last">
    <div class="prefslabel">Color Cycle Speed</div>
    <div id="cpsSlider" x-mojo-element="Slider"></div>
   </div>

  </div>
 </div>

</div>

