"*title:			Dali Clock",
"*mode:			12",
"*datemode:		MM/DD/YY",
"*seconds:		True",
"*cycle:			True",
"*transparent:		False",
"*memory:		Low",
"*foreground:		white",
"*background:		black",
"*visualID:		default",
"XDaliClock*font:	BUILTIN",
