/* xdaliclock - a melting digital clock
 * Copyright (c) 1991-2010 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

#import <UIKit/UIKit.h>

@class DaliClockView;

@interface DaliPhoneAppDelegate : NSObject {
  IBOutlet UIWindow *window;
  IBOutlet DaliClockView *contentView;
}

@property (nonatomic, retain) UIWindow *window;
@property (nonatomic, retain) DaliClockView *contentView;

@end
